/*
  $Id: wkstat.c,v 1.4 1999-11-19 13:08:33-06 annis Exp annis $
  $Source: /u/annis/code/NewMom/kstat/RCS/wkstat.c,v $

  Copyright (c) 1999, 2000 William S. Annis.  All rights reserved.
  This is free software; you can redistribute it and/or modify it
  under the same terms as Perl (the Artistic Licence).  Developed at
  the Department of Biostatistics and Medical Informatics, University
  of Wisconsin, Madison.

  Iterator library interface to libkstat.  Intended to be wrapped by
  SWIG or some other such tool.

  This library is informed by the mechanism DBM(3) uses to let you
  cycle through all values of the database: wkstat_open, (ks_start?),
  wkstat_next, wkstat_close.  The only difference here is that you'll
  need a few more things: a kstat handle and the actual kstat itself,
  type checking, then some basic data retrieval.  More complex data
  access should be handled by the user.

*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <kstat.h>

#include "wkstat.h"


wks_kst_wt
wkstat_open(void)
{
  wks_kst_wt wks;
  
  wks.kctl = kstat_open();
  if (wks.kctl == NULL)
    {
      /* Barf... something ill. */
      perror("wkstat_open:kstat_open");
      exit(1);
    }
  
  /* Leap into the kstat chain. */
  wks.ksp = wks.kctl->kc_chain;

  return wks;
}

/*
  Go back to the head of the chain.  Sensibly, kstat_read handles the
  fetching of data, so subsequent sweeps have new data.
*/
void
wkstat_rewind(wks_kst_wt *wks)
{
  wks->ksp = wks->kctl->kc_chain;
}


/* 
   The return values are to allow the idiom:
   do { something(); } while (wkstat_next(&wkc));
*/
int
wkstat_next(wks_kst_wt *wks)
{
  if (wks->ksp != NULL)
    {
      if (wks->ksp = wks->ksp->ks_next)
        {
          return 1;             /* Data here. */
        }
      else
        {
          return 0;             /* NULL here. */
        }
    }
  else
    {
      /* What are you doing?! */
      return -1;
    }
}

/*
  What's the name of the current kstat.
*/
char *
wkstat_name(wks_kst_wt *wks)
{
  return wks->ksp->ks_name;
}

char *
wkstat_class(wks_kst_wt *wks)
{
  return wks->ksp->ks_class;
}

char *
wkstat_module(wks_kst_wt *wks)
{
  return wks->ksp->ks_module;
}

int
wkstat_instance(wks_kst_wt *wks)
{
  return wks->ksp->ks_instance;
}

int
wkstat_type(wks_kst_wt *wks)
{
  return wks->ksp->ks_type;
}

long
wkstat_created(wks_kst_wt *wks)
{
  return (long)wks->ksp->ks_crtime;
}

long
wkstat_snaptime(wks_kst_wt *wks)
{
  return (long)wks->ksp->ks_snaptime;
}


/* Let the error code percolate */
int
wkstat_close(wks_kst_wt *wks)
{
  return kstat_close(wks->kctl);
}

/* Useful for debugging. */
char *
wkstat_type_as_string(wks_kst_wt *wks)
{
  static char *types[] = {
    "KSTAT_TYPE_RAW",
    "KSTAT_TYPE_NAMED",
    "KSTAT_TYPE_INTR",
    "KSTAT_TYPE_IO",
    "KSTAT_TYPE_TIMER"
  };
  
  return types[wks->ksp->ks_type];
}

  
/*
  Read this element: does a kstat_read.  If the type is NAMED, sets
  up the knamed and knamed_i parts of the wks_kst handle.

  The return code percolates up: -1 is a failure.
*/
kid_t
wkstat_read(wks_kst_wt *wks)
{
  kid_t kid;
  
  kid = kstat_read(wks->kctl, wks->ksp, NULL);
  if (kid > 0 && WKSTAT_NAMED(wks))
    {
      wks->knamed = KSTAT_NAMED_PTR(wks->ksp);
      /* The first thing done to this in wkstat_named_read is an increment. */
      wks->knamed_i = -1;
    }
  /*  else if (kid > 0 && WKSTAT_IO(wks))
    {
      IO -- this can't have multiple elements.  Let the user
      cope with it.
    }
  */
  return kid;
}

/*
  Increment the pointer for knamed, and return the new name/value pair.
  Can be used in a while(wkstat_named_read(&wks)) { sth(); } loop.
*/
kstat_named_t *
wkstat_named_read(wks_kst_wt *wks)
{
  wks->knamed_i++;
  
  if (wks->knamed_i >= wks->ksp->ks_ndata)
    {
      return (kstat_named_t *)NULL;
    }
  else
    {
      return &wks->knamed[wks->knamed_i];
    }
}

/*
  Return a pointer into the IO struct.  The read has taken place
  in wkstat_read!
*/
kstat_io_t *
wkstat_io_read(wks_kst_wt *wks)
{
  return KSTAT_IO_PTR(wks->ksp);
}

/*
  Have we had an update?  0 on no update, -1 on failure and new
  KID when an update happens.
*/
int
wkstat_update(wks_kst_wt *wks)
{
  return (int)kstat_chain_update(wks->kctl);
}

/*
  There are several interesting kstats which are unfortunately kept RAW.
  These must be cast as pointers to something more usable.

  See /usr/include/sys/sysinfo.h
*/

/* Most basic vminfo */
vminfo_t *
wkstat_vminfo_read(wks_kst_wt *wks)
{
  return (vminfo_t *)wks->ksp->ks_data;
}

/* Sysinfo. */
sysinfo_t *
wkstat_sysinfo_read(wks_kst_wt *wks)
{
  return (sysinfo_t *)wks->ksp->ks_data;
}

cpu_stat_t *
wkstat_cpustat_read(wks_kst_wt *wks)
{
  return (cpu_stat_t *)wks->ksp->ks_data;
}

/* Name cache statistics.  See /usr/include/sys/dnlc.h */
struct ncstats *
wkstat_ncstats_read(wks_kst_wt *wks)
{
  return (struct ncstats *)wks->ksp->ks_data;
}


/* EOF */
